/*
 *  Copyright (c) 1992-2012 Matisse Software Inc. All Rights Reserved.
 *
 *  This file (in both binary and source code form) is subject to the
 *  Supplemental License Terms governing use, modification and redistribution
 *  of Sample Code accompanying the Matisse(R) software.
 *
 */

// the required Matisse C++ header
#include "matisseCXX.h"
// the headers for schema classes
#include "Person.h"
#include "Employee.h"
// doing some read/write to files
#include <fstream>

// namespaces containing Matisse classes
using namespace matisse;
using namespace matisse::reflect;

// namespaces containing the Schema classes
using namespace examples::cxx_examples::chaps_4_5;

// A more complete example using classes to modify Matisse objects.
// Uses an image file to stream data in/out of an attribute. Also shows
// how to catch a specific exception.
int main(int ac, char **av)
{
  if (ac < 3)
  {
    std::cerr << "Usage: " << av[0] << " <host> <db>" << std::endl;
    return -1;
  }
  try
  {
    // In this example, there is a direct namespace mapping between 
    // the C++ namespace and the Schema namespace, so the default factory can handle it
    MtDatabase db(av[1], av[2]);

    // open, select and start access to the database
    db.open();
    db.startTransaction();

    // create a new Employee
    Employee& e = Employee::create(db);

    // setting string attributes
    e.setComment("setting values");
    e.setFirstName("John");
    e.setLastName("Jones");

    // setting numbers
    e.setAge(42);

    // setting Date
    matisse::MtTimestamp ts("2002-02-02");
    e.setHireDate(ts);

    // setting Numeric
    matisse::MtNumeric num(2958.33);
    e.setSalary(num);

    // getting
    std::cout << std::endl << e.getComment() << std::endl;
    std::cout << "\tEmployee: " << e.getFirstName() << " "
              << e.getLastName() << std::endl;

    // suppress output if no value set
    // use generated isAgeNull() method to check if value is null
    if (!(e.isAgeNull()))
      std::cout << "\t" << e.getAge() << " years old" << std::endl;
    std::cout << "\tNumber of Dependents: " << e.getDependents()
              << std::endl;
    std::cout << "\tSalary: $" << e.getSalary() << std::endl;
    std::cout << "\tHired on: " << e.getHireDate() << std::endl;

    // changing values (getting and setting)
    e.setDependents(e.getDependents() + 2);
    if (e.getFirstName() == "John")
    {
      matisse::MtNumeric sal = e.getSalary();
      matisse::MtNumeric mul(1.05);
      // operations on matisse::MtNumeric supported
      sal *= mul;
      e.setSalary(sal);
    }
    e.setComment("changing values");

    // Getting again to show effect of removing value
    // similar to above
    std::cout << std::endl << e.getComment() << std::endl;
    std::cout << "\tEmployee: " << e.getFirstName() << " "
              << e.getLastName() << std::endl;
    // use generated isAgeNull() method to check if value is null
    if (!(e.isAgeNull()))
      std::cout << "\t" << e.getAge() << " years old" << std::endl;
    std::cout << "\tNumber of Dependents: " << e.getDependents()
              << std::endl;
    std::cout << "\tSalary: $" << e.getSalary() << std::endl;
    std::cout << "\tHired on: " << e.getHireDate() << std::endl;

    // Blob access:
    // The Photo attribute's type is MtImage which is
    // streamable. Use an array of bytes to read an image file from
    // disk and store into the attribute.

    std::cout << std::endl << "Accessing streamable attribute"
              << std::endl;

    // setting blob
    int bufSize = 15;       // small buff size for demo purposes
    int actualBytes;
    int totalBytes = 0;
    // use auto_ptr to clean up when done
    std::auto_ptr<char> tmpBytes(new char[bufSize]);
 
    // binary file on disk
    std::ifstream is("matisse.gif", std::ios::in | std::ios::binary);
    if (!is)
    {
      std::cerr << "matisse.gif does not exists" << std::endl;
    }
    else
    {
      // initial call to truncate data
      e.setPhotoElements((::MtByte*)tmpBytes.get(), 0,
                         MT_BEGIN_OFFSET, true);
      while (!is.eof())
      {
        actualBytes = (is.read(tmpBytes.get(), bufSize)).gcount();
        e.setPhotoElements((::MtByte*)tmpBytes.get(), actualBytes,
                           MT_CURRENT_OFFSET, true);
        totalBytes += actualBytes;
      }
      is.close();
    }
    std::cout << "transferred " << totalBytes
              << " bytes from matisse.gif to object"
              << std::endl;

    // read blob and store to file, just reverse the process from above
    std::ofstream os("new.gif", std::ios::binary);
    totalBytes = 0;
    if (!os)
    {
      std::cerr << "Failure opening new.gif" << std::endl;
    }
    else
    {
      // reset the steam
      e.getPhotoElements((::MtByte*)tmpBytes.get(), 0, MT_BEGIN_OFFSET);
      do {
        actualBytes = e.getPhotoElements((::MtByte*)tmpBytes.get(),
                                         bufSize, MT_CURRENT_OFFSET);
        os.write(tmpBytes.get(), actualBytes);
        totalBytes += actualBytes;
      } while (actualBytes == bufSize);
      os.close();
    }
    std::cout << "transferred " << totalBytes
              << " bytes from object to new.gif"
              << std::endl;


    db.commit();

    // set read-only access and catch exception

    db.startVersionAccess();

    std::cout << "\nAttempting to modify " << e.getFirstName() + " "
              << e.getLastName() << " salary within read-only access"
              << std::endl;
    try
    {
      matisse::MtNumeric num(3958.40);
      e.setSalary(num);
    }
    catch (MtException& mte)
    {
      std::cout << mte;
      if (mte.getErrorCode() == MATISSE_NOTRANS)
        std::cout << " as expected..." << std::endl;
      else
        std::cout << " but was not expected..." << std::endl;
    }
        
    db.endVersionAccess();

    // remove created object from database
    std::cout << "\nRemoving created object" << std::endl;
    db.startTransaction();
    e.remove();
    db.commit();
        
    db.close();
  }
  catch (MtException &e)
  {
    std::cerr << e << std::endl;
  }
  catch (...)
  {
    std::cerr << "Unknown exception" << std::endl;
  }
  return 0;
}
