/*
 *  EventsExample.cpp
 *
 *  Copyright (c) 1992-2012 Matisse Software, Inc. All Rights Reserved.
 *
 * 
 *  This file (in both binary and source code form) is subject to the
 *  Supplemental License Terms governing use, modification and redistribution
 *  of Sample Code accompanying the Matisse(R) software.
 */

//  Microsoft Visual C++ 
#if defined _MSC_VER
#include <windows.h>
#define SLEEP(t) Sleep((t)*1000)
#else
#include <unistd.h>
#define SLEEP(t) sleep((t))
#endif

// the default Matisse C++ header
#include "matisseCXX.h"


using namespace std;
// the namespaces that contain the Matisse classes
using namespace matisse;

/**
 * This examples illustrates Matisse Event Notification mechansim.
 * The sample application is divided in 2 sections. The first section 
 * is event selection and notification. The second section is event registration 
 * and event handling.
 */

const int TEMPERATURE_CHANGES_EVT = MT_EVENT1;
const int RAINFALL_CHANGES_EVT = MT_EVENT2;
const int HIMIDITY_CHANGES_EVT = MT_EVENT3;
const int WINDSPEED_CHANGES_EVT = MT_EVENT4;

const unsigned int MAX_MEASURES = 12;
static double tempMeasures[] = { 48.7, 52.2, 53.3, 55.6, 58.1, 61.5, 62.7, 63.7, 64.5, 61.0, 54.8, 49.4 };
static double rainfallMeasures[] = { 4.72, 4.15, 3.40, 1.25, 0.54, 0.13, 0.04, 0.09, 0.28, 1.19, 3.31, 3.18 };

bool ChangeInTemperature(int idx)
{
  bool res = true;

  if (idx > 0) {
    double absVal = (tempMeasures[idx] > tempMeasures[idx - 1] ? 
		     tempMeasures[idx] - tempMeasures[idx - 1] : 
		     tempMeasures[idx - 1] - tempMeasures[idx]);
    res = (absVal > 2.5);
  }
  return res;
}

bool ChangeInRainfall(int idx)
{
  bool res = true;

  if (idx > 0) {
    double absVal = ((rainfallMeasures[idx] > rainfallMeasures[idx - 1]) ?
		     (rainfallMeasures[idx] - rainfallMeasures[idx - 1]) :
		     (rainfallMeasures[idx - 1] - rainfallMeasures[idx]));
      res = (absVal > 0.25);
    }
  return res;
}


/** Execute an event selection and notification */
static void NotifyOfChanges(std::string hostname, std::string dbname)
{
  printf("\n\n=========== Notify Of Changes ==========\n");
  try {

    // wait a little ...
    printf("waiting 3 seconds one or more subscribers can be started ...\n");
    SLEEP(3);

    MtDatabase &dbcon = *new MtDatabase(hostname, dbname);

    // Open the connection to the database
    dbcon.open();

    matisse::MtEvent &notifier = *new matisse::MtEvent(dbcon);

    ::MtEvent eventSet;

    for (unsigned int i = 0; i < MAX_MEASURES; i++) {
	eventSet = 0;
	if (ChangeInTemperature(i)) eventSet |= TEMPERATURE_CHANGES_EVT;
	if (ChangeInRainfall(i)) eventSet |= RAINFALL_CHANGES_EVT;
	if (i % 2 == 0) eventSet |= HIMIDITY_CHANGES_EVT;
	if (i % 2 == 1) eventSet |= WINDSPEED_CHANGES_EVT;

	notifier.notify(eventSet);

	printf("Events (#%d) posted:\n", (i+1));
	if ((eventSet & TEMPERATURE_CHANGES_EVT) > 0) printf("Change in temperature\n");
	if ((eventSet & RAINFALL_CHANGES_EVT) > 0) printf("Change in rain fall\n");
	if ((eventSet & HIMIDITY_CHANGES_EVT) > 0) printf("Change in humidity\n");
	if ((eventSet & WINDSPEED_CHANGES_EVT) > 0) printf("Change in wind speed\n");

	// wait a little ...
	printf("waiting 1 sec ...\n");
	SLEEP(1);
      }
    // Close the database connection
    dbcon.close();

  } catch(MtException &e)  {
    fprintf(stderr, "Matisse exception: %s\n", e.getMessage().c_str());
  } catch (...) {
    fprintf(stderr, "unknown exception\n");
  }
}


/** Execute an event selection and notification */
static void SubscribeToChanges(std::string hostname, std::string dbname)
{
  cout << "\n\n=========== Notify Of Changes ==========\n" << endl;
  try {

    MtDatabase &dbcon = *new MtDatabase(hostname, dbname);

    // Open the connection to the database
    dbcon.open();

    matisse::MtEvent &subscriber = *new matisse::MtEvent(dbcon);

    // Subscribe to all 4 events
    ::MtEvent eventSet = TEMPERATURE_CHANGES_EVT |
	RAINFALL_CHANGES_EVT |
	HIMIDITY_CHANGES_EVT |
	WINDSPEED_CHANGES_EVT;
  
    subscriber.subscribe(eventSet);

    cout << "Subscribed to 4 Events:" << endl;
    cout << "- TEMPERATURE_CHANGES" << endl;
    cout << "- RAINFALL_CHANGES" << endl;
    cout << "- HIMIDITY_CHANGES" << endl;
    cout << "- WINDSPEED_CHANGES\n" << endl;

    ::MtEvent triggeredEvents;
    int i = 1;
    while (i < 20) {
	cout << "Wait for Events (#" << i << ") ..." << endl;
      // Wait 1000 ms for events to be triggered 
      // return false if not event is triggered until the timeout is reached
      if (subscriber.wait(1090, &triggeredEvents)) {
	cout << "Events (#" << i << ") triggered:" << endl;
	cout << (((triggeredEvents & TEMPERATURE_CHANGES_EVT) > 0) ? "" : "No ") 
	     << "Change in temperature" << endl;
	cout << (((triggeredEvents & RAINFALL_CHANGES_EVT) > 0) ? "" : "No ")
	     <<  "Change in rain fall" << endl;
	cout << (((triggeredEvents & HIMIDITY_CHANGES_EVT) > 0) ? "" : "No ")
	     << "Change in humidity" << endl;
	cout << (((triggeredEvents & WINDSPEED_CHANGES_EVT) > 0) ? "" : "No ")
	     << "Change in wind speed\n" << endl;
      } else {
	cout << "No Event received after ~1 sec\n" << endl;
      }
      i++;
    }

    cout << "Unsubscribe to 4 Events" << endl;
    // Unsubscribe to all 4 events
    subscriber.unsubscribe();

    // Close the database connection
    dbcon.close();

  } catch(MtException &e)  {
    cerr << "Matisse exception: " << e.getMessage().c_str() << endl;
  } catch (...) {
    cerr << "unknown exception" << endl;
  }
}


int main(int argc, char **argv)
{
  string opNotify = string("N");

  if (argc < 4) {
    fprintf(stderr, "usage: %s <host> <db> N|S\n", argv[0]);
    return 1;
  }

  string hostname = std::string(argv[1]);
  string dbname = std::string(argv[2]);
  string operationType = std::string(argv[3]);

  if (opNotify.compare(operationType) == 0) {
    // event selection and notification
    NotifyOfChanges(hostname, dbname);
  } else {
    // event registration and event handling
    SubscribeToChanges(hostname, dbname);
  }

  return 0;
}
